#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include "debug.h"
#include "gc.h"


#define MMAP_FILE	"/usr/local/etc/tunespersto"	/* persistent store */
#define SPACE_SIZE	16384		/* heap size in bytes */
#define MMAP_LEN	(SPACE_SIZE*2)

void getmem(void);      /* mmap() some memory and assign to the GC ptrs      */

/*
 * Here come the global pointers, I hope this is all of em.
 * Why do I have to bother with the char * and void *'s anyways? I'm 
 * starting to get really annoyed with them. :)
 */

PTR fromspace;	/* The place where objects are copied from by the GC */
PTR tospace;	/* tospace is where objects are copied to/created */
PTR scanptr;	/* The address of the next cell to be scanned */
PTR top;	/* Pointer to the top of the free space in tospace */
PTR bottom;	/* The bottom of the available free space in tospace */
PTR _top;	/* These are used to make sure we know the absolute */
PTR _bottom;	/* start/end of the memory block (just in case) */

int space_size = SPACE_SIZE/4 ; /* number of memory cells */


/* We'll try stacks later */
/*int stack_top = 0;*/  /* The top element of the stack                      */
/*int stack_scan = 0;*/ /* The latest stack element to be scanned            */

char *progname;

int main(int argc, char *argv[]) {
  PTR foo, bar, baz, blah;

  progname = argv[0];     /* I hate this. */
  getmem();

  dump_ptrs(); /* We want to see the allocations, so print the initial ptrs */
  foo = cons(0, 0);dump_ptrs();    /* Lame tests. I definitly need    */
  bar = cons(0, 0);dump_ptrs();    /* some type of scheme interpreter */
  baz = cons(0, 0);dump_ptrs();    /* or something to test with.      */
  blah = cons(0, 0);dump_ptrs();

  rplacd(foo, MEM_of_PTR(bar));
  rplacd(bar, MEM_of_PTR(baz));
  rplacd(baz, MEM_of_PTR(foo));
  dump_ptrs();
  rplaca(foo, MEM_of_PTR(foo));
  rplaca(bar, MEM_of_PTR(foo));
  rplaca(baz, MEM_of_PTR(foo));
  dump_ptrs();

  flip();  /* switch the (from/to)spaces. My mmap file is much too big */
  
  blah = M2P(cdr(foo));
  blah = cons(P2M(foo), P2M(bar));
  blah = cons(P2M(foo), P2M(bar));
  blah = cons(P2M(foo), P2M(bar));

  dump_ptrs();

  return 0;
}

void getmem(void) {
  int fd;

  if((fd = open(MMAP_FILE, O_RDWR)) == -1) {
    perror(progname);
    exit(-1);
  }
  if((_bottom = mmap(NULL, MMAP_LEN, PROT_READ|PROT_WRITE,
		      MAP_SHARED|MAP_FILE, fd, 0)) == (PTR) -1) {
    perror(progname);
    exit(-1);
  }

  _top        = _bottom+MMAP_LEN;
  tospace     = _bottom;
  fromspace   = _bottom + space_size;
  top         = tospace + space_size - 2;
  bottom      = tospace;
  scanptr     = tospace;
}

void dump_ptrs(void) {
  errprintf(
"_bottom:\t%p
_top:\t\t%p
bottom:\t\t%p
fromspace:\t%p
tospace:\t%p
scanptr:\t%p
top:\t\t%p
", _bottom, _top, bottom, fromspace, tospace, scanptr, top);
}

void error(char *error) {
  errprintf(error);
  dump_ptrs();
  exit(-1);
}
